/*
 * Syncdaemon API
 *
 * Authors: Rodrigo Moya <rodrigo.moya@canonical.com>
 *
 * Copyright 2010 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 3, as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranties of
 * MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "config.h"
#include <string.h>
#ifdef HAVE_GDBUS
#include <gio/gio.h>
#else
#include <dbus/dbus-glib.h>
#endif
#include "syncdaemon-daemon.h"
#include "syncdaemon-config-interface.h"

G_DEFINE_TYPE(SyncdaemonDaemon, syncdaemon_daemon, G_TYPE_OBJECT)

struct _SyncdaemonDaemonPrivate {
#ifdef HAVE_GDBUS
	GDBusConnection *bus;
#else
	DBusGConnection *bus;
#endif
	GObject *dbus_proxy;

	/* Interfaces */
	GObject *daemon_interface;
	GHashTable *subinterfaces;

	/* Status */
	gboolean connected;
	gchar *root_dir;
};

/* Signals */
enum {
	CONNECTED_SIGNAL,
	DISCONNECTED_SIGNAL,
	LAST_SIGNAL
};

static guint daemon_signals[LAST_SIGNAL] = { 0, };

static void
syncdaemon_daemon_finalize (GObject *object)
{
	SyncdaemonDaemon *daemon = SYNCDAEMON_DAEMON (object);

	if (daemon->priv != NULL) {
		g_hash_table_destroy (daemon->priv->subinterfaces);

		if (daemon->priv->dbus_proxy != NULL)
			g_object_unref (daemon->priv->dbus_proxy);

		if (daemon->priv->bus != NULL) {
#ifdef HAVE_GDBUS
			g_object_unref (G_OBJECT (daemon->priv->bus));
#else
			dbus_g_connection_unref (daemon->priv->bus);
#endif
		}

		if (daemon->priv->root_dir != NULL)
			g_free (daemon->priv->root_dir);
		        
		g_free (daemon->priv);
	}

	G_OBJECT_CLASS (syncdaemon_daemon_parent_class)->finalize (object);
}

static void
syncdaemon_daemon_class_init (SyncdaemonDaemonClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	object_class->finalize = syncdaemon_daemon_finalize;

	/* Register signals */
	daemon_signals[CONNECTED_SIGNAL] = g_signal_new ("connected",
							 G_TYPE_FROM_CLASS (klass),
							 (GSignalFlags) G_SIGNAL_RUN_LAST,
							 G_STRUCT_OFFSET (SyncdaemonDaemonClass, connected),
							 NULL, NULL,
							 g_cclosure_marshal_VOID__VOID,
							 G_TYPE_NONE, 0);
	daemon_signals[DISCONNECTED_SIGNAL] = g_signal_new ("disconnected",
							    G_TYPE_FROM_CLASS (klass),
							    (GSignalFlags) G_SIGNAL_RUN_LAST,
							    G_STRUCT_OFFSET (SyncdaemonDaemonClass, disconnected),
							    NULL, NULL,
							    g_cclosure_marshal_VOID__VOID,
							    G_TYPE_NONE, 0);
}

static void
setup_daemon_interface (SyncdaemonDaemon *daemon)
{
	daemon->priv->daemon_interface = g_object_new (SYNCDAEMON_TYPE_INTERFACE, NULL);
	syncdaemon_interface_setup_proxy (SYNCDAEMON_INTERFACE (daemon->priv->daemon_interface),
					  "com.ubuntuone.Syncdaemon", "/",
					  "com.ubuntuone.Syncdaemon.SyncDaemon");
}

static void
name_owner_changed_cb (DBusGProxy *proxy,
		       const char *name,
		       const char *old_owner,
		       const char *new_owner,
		       gpointer user_data)
{
	SyncdaemonDaemon *daemon = SYNCDAEMON_DAEMON (user_data);

	if (g_strcmp0 (name, "com.ubuntuone.Syncdaemon") == 0) {
		if (new_owner != NULL && strlen (new_owner) > 0) {
			setup_daemon_interface (daemon);
		} else {
			g_debug ("Syncdaemon service died");
			if (daemon->priv->daemon_interface != NULL) {
				g_object_unref (daemon->priv->daemon_interface);
				daemon->priv->daemon_interface = NULL;
			}

			g_hash_table_remove_all (daemon->priv->subinterfaces);
			daemon->priv->connected = FALSE;
		}
	}
}

static void
syncdaemon_daemon_init (SyncdaemonDaemon *daemon)
{
	GError *error = NULL;

	daemon->priv = g_new0 (SyncdaemonDaemonPrivate, 1);

	daemon->priv->connected = FALSE;
	daemon->priv->subinterfaces = g_hash_table_new_full (g_str_hash, g_str_equal,
							     g_free, g_object_unref);

	/* Initialize DBus */
#ifdef HAVE_GDBUS
	daemon->priv->bus = g_bus_get_sync (G_BUS_TYPE_SESSION, NULL, &error);
#else
	daemon->priv->bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
#endif
	if (error != NULL) {
		g_warning ("Couldn't get session bus: %s", error->message);
		g_error_free (error);
		return;
	}

	/* Listen to DBus for syncdaemon restarts */
	daemon->priv->dbus_proxy = (GObject *) dbus_g_proxy_new_for_name (daemon->priv->bus,
									  "org.freedesktop.DBus",
									  "/org/freedesktop/DBus",
									  "org.freedesktop.DBus");
	dbus_g_proxy_add_signal (DBUS_G_PROXY (daemon->priv->dbus_proxy), "NameOwnerChanged",
				 G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING,
				 G_TYPE_INVALID);
	dbus_g_proxy_connect_signal (DBUS_G_PROXY (daemon->priv->dbus_proxy),
				     "NameOwnerChanged",
				     G_CALLBACK (name_owner_changed_cb),
				     daemon, NULL);

	/* Get the main interface */
	setup_daemon_interface (daemon);
}

/**
 * syncdaemon_daemon_new:
 *
 * Create a new #SyncdaemonDaemon object, which provides access to the
 * Syncdaemon daemon.
 *
 * Return value: A new #SyncdaemonDaemon object.
 */
SyncdaemonDaemon *
syncdaemon_daemon_new (void)
{
	return g_object_new (SYNCDAEMON_TYPE_DAEMON, NULL);
}

static void
connect_response_cb (DBusGProxy *proxy, DBusGProxyCall *call_id, gpointer user_data)
{
	GError *error = NULL;
	SyncdaemonDaemon *daemon = (SyncdaemonDaemon *) user_data;

	if (dbus_g_proxy_end_call (proxy, call_id, &error, G_TYPE_INVALID)) {
		daemon->priv->connected = TRUE;
		g_signal_emit (daemon, daemon_signals[CONNECTED_SIGNAL], 0);
	} else {
		g_warning ("Syncdaemon cannot connect: %s", error->message);
		g_error_free (error);
	}
}

/**
 * syncdaemon_daemon_connect:
 */
gboolean
syncdaemon_daemon_connect (SyncdaemonDaemon *daemon)
{
	GObject *proxy;

	g_return_val_if_fail (SYNCDAEMON_IS_DAEMON (daemon), FALSE);

	proxy = syncdaemon_interface_get_proxy_object (SYNCDAEMON_INTERFACE (daemon->priv->daemon_interface));
	if (proxy != NULL) {
#ifdef HAVE_GDBUS
#else
		if (!dbus_g_proxy_begin_call (DBUS_G_PROXY (proxy), "connect",
					      connect_response_cb, daemon, NULL,
					      G_TYPE_INVALID)) {
			g_warning ("Call to 'connect' method failed");

			return FALSE;
		}
#endif
	}

	return TRUE;
}

static void
disconnect_response_cb (DBusGProxy *proxy, DBusGProxyCall *call_id, gpointer user_data)
{
	GError *error = NULL;
	SyncdaemonDaemon *daemon = (SyncdaemonDaemon *) user_data;

	if (dbus_g_proxy_end_call (proxy, call_id, &error, G_TYPE_INVALID)) {
		daemon->priv->connected = FALSE;
		g_signal_emit (daemon, daemon_signals[DISCONNECTED_SIGNAL], 0);
	} else {
		g_warning ("Syncdaemon cannot disconnect: %s", error->message);
		g_error_free (error);
	}
}

/**
 * syncdaemon_daemon_interface_disconnect:
 */
gboolean
syncdaemon_daemon_disconnect (SyncdaemonDaemon *daemon)
{
	GObject *proxy;

	g_return_val_if_fail (SYNCDAEMON_IS_DAEMON (daemon), FALSE);

	proxy = syncdaemon_interface_get_proxy_object (SYNCDAEMON_INTERFACE (daemon->priv->daemon_interface));
	if (proxy != NULL) {
#ifdef HAVE_GDBUS
#else
		if (!dbus_g_proxy_begin_call (DBUS_G_PROXY (proxy), "disconnect",
					      disconnect_response_cb, daemon, NULL,
					      G_TYPE_INVALID)) {
			g_warning ("Call to 'disconnect' method failed");

			return FALSE;
		}
#endif
	}

	return TRUE;
}

/**
 * syncdaemon_daemon_interface_quit:
 */
gboolean
syncdaemon_daemon_quit (SyncdaemonDaemon *daemon)
{
	GObject *proxy;

	g_return_val_if_fail (SYNCDAEMON_IS_DAEMON (daemon), FALSE);

	proxy = syncdaemon_interface_get_proxy_object (SYNCDAEMON_INTERFACE (daemon->priv->daemon_interface));
	if (proxy != NULL) {
		GError *error = NULL;

#ifdef HAVE_GDBUS
#else
		if (!dbus_g_proxy_call (DBUS_G_PROXY (proxy), "quit", &error,
					G_TYPE_INVALID,
					G_TYPE_INVALID)) {
			g_warning ("Could not quit syncdaemon: %s", error->message);
			g_error_free (error);

			return FALSE;
		}
#endif
	}

	return TRUE;
}

/**
 * syncdaemon_daemon_get_root_dir:
 */
const gchar *
syncdaemon_daemon_get_root_dir (SyncdaemonDaemon *daemon)
{
	GObject *proxy;

	g_return_val_if_fail (SYNCDAEMON_IS_DAEMON (daemon), NULL);

	if (daemon->priv->root_dir != NULL)
		return (const gchar *) daemon->priv->root_dir;

	proxy = syncdaemon_interface_get_proxy_object (SYNCDAEMON_INTERFACE (daemon->priv->daemon_interface));
	if (proxy != NULL) {
		gchar *new_root_dir;
		GError *error = NULL;

#ifdef HAVE_GDBUS
#else
		if (!dbus_g_proxy_call (DBUS_G_PROXY (proxy), "get_root_dir", &error,
					G_TYPE_INVALID,
					G_TYPE_STRING, &new_root_dir,
					G_TYPE_INVALID)) {
			g_warning ("Could not get syncdaemon's root dir: %s", error->message);
			g_error_free (error);

			return NULL;
		}
#endif

		daemon->priv->root_dir = new_root_dir;
	}

	return NULL;
}

static SyncdaemonInterface *
get_interface (SyncdaemonDaemon *daemon, const gchar *path, SyncdaemonInterface * (* new_func) (void))
{
	SyncdaemonInterface *interface;

	interface = g_hash_table_lookup (daemon->priv->subinterfaces, path);
	if (interface == NULL) {
		interface = new_func ();
		g_hash_table_insert (daemon->priv->subinterfaces, g_strdup (path), interface);
	}

	return interface;
}

/**
 * syncdaemon_daemon_get_config_interface:
 */
SyncdaemonInterface *
syncdaemon_daemon_get_config_interface (SyncdaemonDaemon *daemon)
{
	g_return_val_if_fail (SYNCDAEMON_IS_DAEMON (daemon), NULL);

	return get_interface (daemon, "/config", syncdaemon_config_interface_new);
}

/**
 * syncdaemon_daemon_get_status_interface:
 */
SyncdaemonInterface *
syncdaemon_daemon_get_status_interface (SyncdaemonDaemon *daemon)
{
	g_return_val_if_fail (SYNCDAEMON_IS_DAEMON (daemon), NULL);
}
